@@ -40,16 +40,21 @@ def CAPS(addrspace):
     Translate an address space constant into a capability bitmask.'''
     return C.ADDRXLAT_CAPS(addrspace)
 
+def pte_format_name(fmt):
+    '''pte_format_name(fmt) -> name
+
+    Return the name of a PTE format constant.'''
+    return utils.to_unicode(ffi.string(C.addrxlat_pte_format_name(fmt)))
+
+def pte_format(name):
+    '''pte_format(name) -> fmt
+
+    Return the PTE format constant by name.'''
+    return C.addrxlat_pte_format(utils.to_bytes(name))
+
 def pteval_shift(fmt):
     '''pteval_shift(fmt) -> bit shift
 
     Get the pteval shift for a PTE format.
     See PTE_xxx for valid values of fmt.'''
     return C.addrxlat_pteval_shift(fmt)
-
-def sym_argc(type):
-    '''sym_argc(type) -> number of arguments
-
-    Get the number of arguments for a given type of symbolic info.
-    See SYM_xxx for valid values of type.'''
-    return C.addrxlat_sym_argc(type)
